from __future__ import absolute_import, division, print_function, unicode_literals
from .util import ListDict, deprecated, abstractmethod
from collections import deque
from .logger import logger
from .schedulers import NaiveScheduler, LinearizedScheduler
import networkx as nx
import functools
import keyword
from warnings import warn
import future
from future.builtins import super

__all__ = "Component", "Actor", "Workflow", "Composite", "draw_graph"


class NoValue(object):
    """A unique no value object

    Note that we cannot use None as this can be used by users
    """

    def __init__(self):
        raise Exception('NoValue cannot be instantiated')


def has_value(value):
    """Returns True for any value except NoValue
    """
    return value is not NoValue


class Component(object):
    """Base WOWP component class
    """

    def __init__(self, name=None):
        if name is None:
            name = self.__class__.__name__.lower()
        assert is_valid_componenet_name(name)
        self.name = name
        self._inports = Ports(InPort, self)
        self._outports = Ports(OutPort, self)

    def get_run_args(self):
        """
        Prepare arguments (args, kwargs) for the run method

        The default behaviour is to put values from
        all connected input ports to kwargs.

        :return: args, kwargs
        :rtype: tuple
        """
        args = ()
        kwargs = {
            port.name: port.pop()
            for port in self.inports if port.isconnected()
            }
        return args, kwargs

    @classmethod
    def run(cls, *args, **kwargs):
        """This is a virtual method
        """
        raise NotImplementedError('Calling a virtual method')

    def can_run(self):
        # print("on_input", all(not port.isempty() for port in self.inports))
        return not any(port.isempty() for port in self.inports)

    @property
    def inports(self):
        return self._inports

    @inports.setter
    def inports(self, value):
        raise TypeError('Cannot set inports directly')

    @inports.deleter
    def inports(self):
        raise TypeError('Cannot delete inports directly')

    @property
    def outports(self):
        return self._outports

    @outports.setter
    def outports(self, value):
        raise TypeError('Cannot set outports directly')

    @outports.deleter
    def outports(self):
        raise TypeError('Cannot delete outports directly')

    def put_outputs(self, **kwargs):
        """Put outputs to ports, specified by port_name, value pairs
        """
        for port_name, value in kwargs.items():
            self.outports[port_name].put(value)

    @property
    def graph(self):
        """Construct NetworX call graph
        """
        return build_nx_graph(self)

    # TODO create workflow (composite) from actor's connections

    def get_workflow(self, name=None):
        """Creates a workflow form actor's connections
        """

        graph = self.graph
        leaves_out = [n for n, d in graph.out_degree_iter() if d == 0]
        leaves_in = [n for n, d in graph.in_degree_iter() if d == 0]

        workflow = Workflow(name=name)

        for node in (graph.node[n] for n in leaves_in):
            if isinstance(node['ref'], Component):
                warn('Component without any input: {} ({})'.format(
                    node['ref'].name, node['ref']))
            elif isinstance(node['ref'], InPort):
                workflow.add_inport(node['ref'])
            else:
                raise Exception('{} cannot be an input port'.format(node['ref'
                                                                    ]))

        for node in (graph.node[n] for n in leaves_out):
            if isinstance(node['ref'], Component):
                if not node['ref'].system_actor:
                    # system actors can be designed not to have outports, e.g. Sink
                    warn('Component without any output: {} ({})'.format(
                        node['ref'].name, node['ref']))
            elif isinstance(node['ref'], OutPort):
                workflow.add_outport(node['ref'])
            else:
                raise Exception('{} cannot be an output port'.format(node['ref'
                                                                    ]))

        return workflow


class Actor(Component):
    """Actor class
    """

    def __call__(self, **kwargs):
        """
        Run the component with input ports filled from keyword arguments.

        :param kwargs: input ports values
        :return: output port(s) value(s)
        :rtype: dict for multiple ports
        """

        for inport in self.inports:
            if inport.name in kwargs:
                inport.buffer.appendleft(kwargs[inport.name])
        # run the actor
        args, kwargs = self.get_run_args()
        res = self.run(*args, **kwargs)
        return res

    @property
    def system_actor(self):
        return getattr(self, '_system_actor', False)


class Composite(Component):
    """Composite = a group of actors
    """

    def __init__(self, name=None, scheduler=None):
        super().__init__(name=name)
        self.scheduler = scheduler

    def __call__(self, scheduler=None, **kwargs):
        """
        Run the component with input ports filled from keyword arguments.

        Args:
            scheduler: execution scheduler (default LinearizedScheduler)
            kwargs: input ports values

        Returns:
            output port(s) value(s)
        """

        if self.scheduler is not None:
            # prefer self.scheduler
            scheduler = self.scheduler
        if scheduler is None:
            # default scheduler for __call__
            scheduler = LinearizedScheduler()

        scheduler.run_workflow(self, **kwargs)
        res = {port.name: port.pop_all() for port in self.outports}
        return res

    def add_inport(self, inport):
        self._inports[inport.name] = inport

    def add_outport(self, outport):
        self._outports[outport.name] = outport


class Workflow(Composite):
    """Workflow class
    """
    pass


class Ports(object):
    """Port collection
    """

    def __init__(self, port_class, owner):
        # TODO port_class can differ for individual ports
        self._ports = ListDict()
        # port class is used to create new ports
        self._port_class = port_class
        self._owner = owner

    def isempty(self):
        return bool(self._ports)

    def __len__(self):
        return len(self._ports)

    def __contains__(self, item):
        return item in self._ports

    def __iter__(self):
        # TODO this might not be intuitive
        return iter(self._ports.values())

    def items(self):
        return self._ports.items()

    def values(self):
        return self._ports.values()

    def __new_port(self, name, **kwargs):
        return self._port_class(name=name, owner=self._owner, **kwargs)

    def __getitem__(self, item):
        """

        :rtype: Port
        """
        # TODO add security
        return self._ports[item]

    def __getattr__(self, item):
        # TODO add security
        return self._ports[item]

    def __setitem__(self, key, value):
        # must be implemented for +=, -= operators
        # TODO add security
        self._ports[key] = value

    def __str__(self):
        return "Ports: [" + ", ".join(self.keys()) + "]"

    def insert_after(self, existing_port_name, new_port_name, replace_existing=False):
        if not replace_existing and new_port_name in self._ports:
            raise Exception('Port {} already exists'.format(new_port_name))
        self._ports.insert_after(
            existing_port_name,
            (new_port_name, self.__new_port(new_port_name)))

    def append(self, new_port_name, replace_existing=False, **kwargs):
        if not replace_existing and new_port_name in self._ports:
            raise Exception('Port {} already exists'.format(new_port_name))
        self._ports[new_port_name] = self.__new_port(new_port_name, **kwargs)

    def keys(self):
        return list(self._ports.keys())

    def at(self, index):
        """Get port by number."""
        key = self.keys()[index]
        return self[key]


class Port(object):
    """Represents a single input/output actor port
    """

    def __init__(self, name, owner, persistent=False, default=NoValue):
        assert is_valid_port_name(name)
        self.name = name
        self.owner = owner
        self.persistent = persistent
        self.buffer = deque()
        self._default = default
        self._connections = []
        self._last_value = NoValue

    @property
    def default(self):
        if has_value(self._default):
            return self._default
        else:
            raise AttributeError('No default value specified')

    @default.setter
    def default(self, value):
        self._default = value

    @default.deleter
    def default(self):
        self._default = NoValue

    @property
    def connections(self):
        return self._connections

    def is_connected_to(self, other):
        return self in other.connections

    def connect(self, other):
        if isinstance(self, OutPort):
            assert isinstance(other, InPort)
        else:
            assert isinstance(other, OutPort)
        if other not in self._connections:
            # cannot use connect as it created an infinite recursion
            self._connections.append(other)
            # TODO this creates a circular reference - is it a good idea?
            other._connections.append(self)
        else:
            logger.warn('connecting an already connected actor {}'.format(
                other))

    @deprecated
    def __bool__(self):
        """True if the port buffer is not empty
        """
        return not self.isempty()

    def isempty(self):
        """True if the port buffer is empty
        """
        if self.buffer or has_value(self._default) or (
                    self.persistent and has_value(self._last_value)):
            return False
        else:
            return True

    def isconnected(self):
        return bool(self._connections)

    def disconnect(self, other):
        if other not in self._connections:
            logger.warn('actor {} not currently connected'.format(other))
        else:
            other._connections.remove(self)
            self._connections.remove(other)

    def pop(self):
        """Get single input
        """
        res = NoValue
        if self.buffer:
            # input item is in the buffer
            res = self.buffer.popleft()
            if self.persistent:
                self._last_value = res
        elif self.persistent and has_value(self._last_value):
            # persistent port, last value exists
            res = self._last_value
        elif has_value(self._default):
            # port with default value
            res = self._default
        # chack whether any result value is available
        if has_value(res):
            return res
        else:
            raise IndexError('Port buffer is empty')

    def pop_all(self):
        """Get all values
        """
        values = self.buffer
        self.buffer = deque()
        return values

    @abstractmethod
    def put(self, value):
        pass


class OutPort(Port):
    """A single, named output port
    """

    def put(self, value):
        """Put output value

        Value is sent to connected ports (or stored if not connected)
        """
        self.buffer.append(value)


class InPort(Port):
    """A single, named input port
    """

    def __iadd__(self, other):
        self.connect(other)
        # self must be returned because __setattr__ or __setitem__ is finally used
        return self

    def __isub__(self, other):
        self.disconnect(other)
        # self must be returned because __setattr__ or __setitem__ is finally used
        return self

    def put(self, value):
        """Put single input

        :rtype: bool
        :return: Whether the actor is ready to perform
        """
        self.buffer.append(value)
        return self.owner.can_run()


def is_valid_port_name(name):
    """Validate port name
    """
    if future.utils.isidentifier(name) and not keyword.iskeyword(name):
        return True
    else:
        return False


def is_valid_componenet_name(name):
    """Validate actor name
    """
    # so far no restrictions
    return True


def build_nx_graph(actor):
    """Create graph with all actors + ports as nodes.

    It walks over all connections.

    Prerequisities:
    * networkx package
    """
    actors = []
    edges = []
    ports = []
    graph = nx.DiGraph()

    def _get_name(obj):
        return str(hash(obj))

    def _add_actor_node(actor):
        attrs = {"fontsize": "12", "color": "#0093d0"}
        # TODO use wekreds for ref
        graph.add_node(_get_name(actor), type='actor', ref=actor, label=actor.name,
                       shape="box", **attrs)
        actors.append(actor)

    def _walk_node(actor):
        name = _get_name(actor)
        if actor not in actors:
            _add_actor_node(actor)
        for port in actor.outports:
            if port not in ports:
                attrs = {}
                if not port.connections:
                    # terminal node
                    attrs["style"] = "filled"
                    attrs["color"] = "#ef4135"
                else:
                    attrs["color"] = "#ffe28a"
                graph.add_node(_get_name(port), type='port', ref=port,
                               label=port.name, **attrs)
                graph.add_edge(name, _get_name(port))
            for other in port.connections:
                if (port, other) not in edges:
                    edges.append((port, other))
                    _walk_node(other.owner)
                    graph.add_edge(_get_name(port), _get_name(other))
        for port in actor.inports:
            if port not in ports:
                attrs = {}
                if not port.connections:
                    # terminal node
                    attrs["style"] = "filled"
                    attrs["color"] = "#ffffff"
                else:
                    attrs["color"] = "#9ed8f5"
                graph.add_node(_get_name(port), type='port', ref=port,
                               label=port.name, **attrs)
                graph.add_edge(_get_name(port), name, )
                ports.append(port)
            for other in port.connections:
                if (other, port) not in edges:
                    _walk_node(other.owner)

    _walk_node(actor)
    return graph


def draw_graph(graph, layout='spectral', with_labels=True, node_size=500,
               pos_kwargs=None, draw_kwargs=None):
    """Draw a workflow graph using NetworkX
    """

    kwargs = {}
    if pos_kwargs is not None:
        kwargs.update(pos_kwargs)
    if layout == 'spectral':
        lfunc = functools.partial(nx.spectral_layout, **kwargs)
    else:
        raise ValueError('{} layout not supported'.format(layout))
    # get colors and labels
    colors = [graph.node[n].get('color', '#ffffff') for n in graph.nodes_iter()
              ]
    shapes = [graph.node[n].get('shape', 'o') for n in graph.nodes_iter()]
    labels = {n: graph.node[n].get('label', '') for n in graph.nodes_iter()}
    pos = lfunc(graph)
    nx.draw_networkx(graph,
                     pos=pos,
                     with_labels=with_labels,
                     labels=labels,
                     node_color=colors,
                     node_size=node_size)



"""OMFIT actors
"""

from wowp.components import Actor
import six
import os
from wowp.logger import logger


def _shell_run(command,
               workdir,
               files_in=(),
               files_out=(),
               timeout=None,
               cleanup=False,
               shell='/bin/bash',
               print_output=True,
               binary_mode=False):
    import subprocess
    import tempfile
    import sys
    from tempfile import mkdtemp
    from glob import glob
    import time

    if workdir is None:
        # TODO handle creating new workdirs
        raise NotImplementedError()

    tmpdir = mkdtemp(dir=workdir)
    print('tmpdir: {}'.format(tmpdir))

    try:
        # put everything into try - finally to clean up tmpdir
        try:
            # link all files in workdir --> should be a cheap and safe operation
            # TODO os.link available only on Unix in 2.7 (the scipt below as well anyway)
            # link input files
            if isinstance(files_in, six.string_types):
                files_in = (files_in, )
            for glob_in in files_in:
                print('glob_in: {}'.format(glob_in))
                if isinstance(glob_in, six.string_types):
                    # use glob in this case, do not rename files
                    # for in_file_name in glob(os.path.join(workdir, glob_in)):
                    for in_file_name in glob(glob_in):
                        print('in_file_name: {}'.format(in_file_name))
                        # source = os.path.join(workdir, os.path.basename(in_file_name))
                        source = os.path.abspath(in_file_name)
                        target = os.path.join(tmpdir, os.path.basename(in_file_name))
                        print('link({}, {})'.format(source, target))
                        os.link(source, target)
                else:
                    # do not use glob if (source, target) filenames are provided
                    # source = os.path.join(workdir, os.path.basename(glob_in[0]))
                    source = os.path.abspath(glob_in[0])
                    target = os.path.join(tmpdir, os.path.basename(glob_in[1]))
                    print('link("{}", "{}")'.format(source, target))
                    os.link(source, target)

        except Exception:
            logger.error('Error linking input files')
            raise

        # construct the full command
        full_command = ('cd {}\n{}\n').format(tmpdir, command)

        if print_output:
            print('run command:\n{}'.format(full_command))

        if binary_mode:
            mode = "w+b"
        else:
            mode = "w+t"

        with tempfile.NamedTemporaryFile(mode=mode,
                                         prefix='stdout',
                                         dir=tmpdir) as fout, tempfile.NamedTemporaryFile(
                                             mode=mode,
                                             prefix='stderr',
                                             dir=tmpdir) as ferr:

            if not isinstance(shell, six.string_types):
                executable = None
            else:
                executable = shell

            if timeout is None:
                result = subprocess.call(full_command,
                                         stdout=fout,
                                         stderr=ferr,
                                         executable=executable,
                                         shell=True)
            else:
                maxtime = time.time() + timeout
                proc = subprocess.Popen(full_command,
                                        executable=executable,
                                        stdout=fout,
                                        stderr=ferr,
                                        shell=True)
                while proc.poll() is None:
                    time.sleep(timeout * 0.01)
                    if time.time() > maxtime:
                        proc.terminate()
                        if print_output:
                            fout.seek(0)
                            ferr.seek(0)
                            cout = fout.read()
                            cerr = ferr.read()
                            sys.stdout.write(cout)
                            sys.stderr.write(cerr)
                        raise Exception('time out expired in {}'.format(full_command))

                result = proc.returncode

            fout.seek(0)
            ferr.seek(0)
            cout = fout.read()
            cerr = ferr.read()

        if print_output:
            sys.stdout.write(cout)
            sys.stderr.write(cerr)

    finally:
        # cleanup the temp directory
        if cleanup:
            for fname in os.listdir(tmpdir):
                if fname not in files_out:
                    os.remove(os.path.join(tmpdir, fname))

    res = {'ret': result, 'stdout': cout, 'stderr': cerr, 'tmpdir': tmpdir}

    return res


class FileCommand(Actor):
    """Use shell command to process files

    Args:
        name (str): the actor name
        command (str): the command to be executed
        input_files: a list of (file_name, port_name) mappings
        output_files: a list of (file_name, port_name) mappings
        workdir (str): where temporary directories will be created
        shell_res: add shell_res output port with shell results and text outputs
        single_out: join outputs into a single dict
        shell(str): the shell path [/bin/bash]
        print_output (bool): print the std out/err [True]
        timeout (float): maximum run time for the executable
        cleanup (bool): cleanup temporary directories
        raise_error (bool): raise and exception in case of an error in the shell process [True]
    """

    def __init__(self,
                 name,
                 command,
                 input_files=(),
                 output_files=(),
                 workdir=None,
                 shell_res=False,
                 single_out=False,
                 shell='/bin/bash',
                 print_output=True,
                 timeout=None,
                 cleanup=False,
                 raise_error=True):
        super(FileCommand, self).__init__(name=name)

        # use input and output file names as ports
        if isinstance(input_files, six.string_types):
            input_files = (input_files, )
        if isinstance(output_files, six.string_types):
            output_files = (output_files, )
        # map of port name -> file name
        self.inports_map = {}
        self.outports_map = {}
        for in_file in input_files:
            if isinstance(in_file, six.string_types):
                file_name, port_name = in_file, in_file
            else:
                file_name, port_name = in_file
            self.inports_map[port_name] = file_name
            self.inports.append(port_name)
        if single_out:
            self.outports.append('out')
        for out_file in output_files:
            if isinstance(out_file, six.string_types):
                file_name, port_name = out_file, out_file
            else:
                file_name, port_name = out_file
            self.outports_map[port_name] = file_name
            if not single_out:
                self.outports.append(port_name)

        if shell_res and not single_out:
            self.outports.append('shell_res')

        self.command = command
        self.workdir = workdir
        self.single_out = single_out
        self.shell = shell
        self.timeout = timeout
        self.cleanup = bool(cleanup)
        self.shell_res = bool(shell_res)
        self.raise_error = bool(raise_error)

    def get_run_args(self):
        # get the input file names
        files_in = []
        files_out = []
        # for inport in self.inports:
        #     files_in.append((os.path.abspath(inport.pop()), self.inports_map[inport.name]))
        # for outport in self.outports:
        #     files_out.append(self.outports_map[outport.name])
        for port_name, file_name in self.inports_map.items():
            files_in.append((os.path.abspath(self.inports[port_name].pop()), file_name))
        for port_name, file_name in self.outports_map.items():
            files_out.append(file_name)
        args = ()
        kwargs = {'files_in': files_in,
                  'outports_map': self.outports_map,
                  'workdir': self.workdir,
                  'command': self.command,
                  'shell': self.shell,
                  'single_out': self.single_out,
                  'timeout': self.timeout,
                  'raise_error': self.raise_error,
                  'shell_res': self.shell_res,
                  'cleanup': self.cleanup}

        return args, kwargs

    @staticmethod
    def run(*args, **kwargs):

        shell_res = _shell_run(kwargs['command'],
                               kwargs['workdir'],
                               files_in=kwargs['files_in'],
                               files_out=kwargs['outports_map'].values(),
                               timeout=kwargs['timeout'],
                               shell=kwargs['shell'],
                               print_output=True,
                               cleanup=kwargs['cleanup'],
                               binary_mode=False)

        # possibly look at shell_res here (for error etc)

        if shell_res['ret'] != 0:
            if kwargs['raise_error']:
                raise Exception('error running toq:\n{}\n{}'.format(shell_res['stdout'],
                                                                    shell_res['stderr']))

        # puth output file names into output ports
        res = {}

        for port_name, file_name in kwargs['outports_map'].items():
            res[port_name] = os.path.join(shell_res['tmpdir'], file_name)
        if kwargs['shell_res']:
            res['shell_res'] = shell_res
        if kwargs['single_out']:
            res = {'out': res}
        return res


def how_to_test():
    from tempfile import mkdtemp
    workdir = mkdtemp()
    print('running in {}'.format(workdir))
    input_file_name_tmp = 'my_input_file_tmp.txt'
    input_file_name = 'my_input_file.txt'
    input_port_name = 'my_input'
    output_file_name = 'my_output_file.txt'
    output_port_name = 'my_output'
    open(
        os.path.join(workdir, input_file_name_tmp),
        'w').write('This is my input file :-P')

    from wowp.actors.omfit import FileCommand

    command = 'echo "I am TOQ" > {output}; cat {input} >> {output}; sleep 1'.format(
        input=input_file_name,
        output=output_file_name)

    toq = FileCommand('toq_actor',
                      command=command,
                      input_files=((input_file_name, input_port_name), ),
                      output_files=((output_file_name, output_port_name), ),
                      workdir=workdir,
                      timeout=None,
                      cleanup=False)

    kwargs = {input_port_name: os.path.join(workdir, input_file_name_tmp)}
    res = toq(**kwargs)
    print('output file content:')
    print(open(res[output_port_name]).read())



from __future__ import absolute_import, division, print_function, unicode_literals

import inspect
from collections import deque
import threading
import warnings
import wowp.components
import time
import datetime
import six
import traceback
from .logger import logger
import os
import random

try:
    from ipyparallel import Client, RemoteError
except ImportError:
    warnings.warn(
        'ipyparallel not installed: IPyClusterScheduler cannot be used')
try:
    from mpi4py import MPI
except ImportError:
    warnings.warn(
        'mpi4py not installed: mpi4py cannot be used')
from .util import MPI_TAGS, loads, dumps
import itertools


__all__ = [
    "NaiveScheduler",
    "LinearizedScheduler",
    "ThreadedScheduler",
    "IPyClusterScheduler",
    "MultiprocessingExecutor",
    "MultiIpyClusterScheduler",
    "FuturesScheduler"]

class _ActorRunner(object):
    """Base class for objects that run actors and process their results.

    It is ok, if the runner is a scheduler at the same time. The separation
    of concepts exists only for the cases when a scheduler needs to run
    actors in parallel (such as ThreadedScheduler).
    """

    def on_outport_put_value(self, outport):
        '''
        Propagates values put into an output port.

        Must be called after outport.put
        :param outport: output port
        :return: None
        '''
        if outport.connections:
            value = outport.pop()
            for inport in outport.connections:
                self.put_value(inport, value)

    def run_actor(self, actor):
        # print("Run actor")
        # TODO replace by an attribute / method call
        if isinstance(actor, wowp.components.Composite):
            self.run_workflow(actor)
        else:
            actor.scheduler = self
            args, kwargs = actor.get_run_args()
            result = actor.run(*args, **kwargs)
            # print("Result: ", result)
            if not result:
                return
            else:
                out_names = actor.outports.keys()
                if not hasattr(result, 'items'):
                    raise ValueError('The execute method must return '
                                     'a dict-like object with items method')
                for name, value in result.items():
                    if name in out_names:
                        outport = actor.outports[name]
                        outport.put(value)
                        self.on_outport_put_value(outport)
                    else:
                        raise ValueError("{} not in output ports".format(name))

    def run_workflow(self, workflow, **kwargs):
        inport_names = tuple(port.name for port in workflow.inports)
        if workflow.scheduler is not None:
            # TODO this seems a bit strange
            scheduler = workflow.scheduler
        else:
            scheduler = self
        for key, value in kwargs.items():
            if key not in inport_names:
                raise ValueError('{} is not an inport name'.format(key))
            inport = workflow.inports[key]
            # put values to connected ports
            scheduler.put_value(inport, kwargs[inport.name])
        # TODO can this be run inside self.execute itsef?
        scheduler.execute()

    def reset(self):
        """Reset the scheduler
        """
        # by default, this method does nothing
        pass

    def shutdown(self):
        pass

    # def __del__(self):
    #     self.shutdown()
    #     super(_ActorRunner, self).__del__()


class NaiveScheduler(_ActorRunner):
    """Scheduler that directly calls connected actors.

    Problem: recursion quickly ends in full call stack.
    """

    def copy(self):
        return self

    def put_value(self, in_port, value):
        should_run = in_port.put(value)
        if should_run:
            self.run_actor(in_port.owner)

    def execute(self):
        pass


class LinearizedScheduler(_ActorRunner):
    """Scheduler that stacks all inputs in a queue and executes them in FIFO order."""

    def __init__(self):
        self.execution_queue = deque()

    def copy(self):
        return self.__class__()

    def put_value(self, in_port, value):
        self.execution_queue.appendleft((in_port, value))

    def execute(self):
        while self.execution_queue:
            in_port, value = self.execution_queue.pop()
            should_run = in_port.put(value)
            if should_run:
                self.run_actor(in_port.owner)


class DistributedExecutor(object):
    """Executes jobs using distributed

    Args:
        uris: one or more dexecuter URI's (str or list)
        min_engines (int): minimum number of engines
        timeout(float): time to wait for engines
    """

    def __init__(self, uris, min_engines=1, timeout=60):
        from distributed import Executor
        if isinstance(uris, six.string_types):
            uris = (uris, )
        self._clients = [Executor(addr) for addr in uris]
        self._current_cli = 0
        # TODO assure

    def _rotate_client(self):
        # TODO pick the first (most) empty one
        current = self._current_cli
        self._current_cli = (self._current_cli + 1) % len(self._clients)
        return self._clients[current]

    def submit(self, func, *args, **kwargs):
        """Submit a function: func(*args, **kwargs) and return a FutureJob.
        """
        cli = self._rotate_client()
        job = cli.submit(func, *args, **kwargs)
        return FutureJob(job)


class MultiprocessingExecutor(object):
    """Executes jobs in local subprocesses using concurrent.futures
    """

    def __init__(self, processes):
        from concurrent.futures import ProcessPoolExecutor
        self._pool = ProcessPoolExecutor(max_workers=processes)

    def submit(self, func, *args, **kwargs):
        """Submit a function: func(*args, **kwargs) and return a FutureJob.
        """
        job = self._pool.submit(func, *args, **kwargs)
        return FutureJob(job)


def mpi_worker():
    """Start a single MPI worker node
    """
    from mpi4py import MPI
    from wowp.util import MPI_TAGS, loads, dumps

    comm = MPI.COMM_WORLD   # get MPI communicator object
    rank = comm.rank        # rank of this process
    status = MPI.Status()   # get MPI status object

    # say hello to the master
    name = MPI.Get_processor_name()
    print("I am a WOW:-P MPI worker with rank %d on %s." % (rank, name))
    msg = {"rank": rank, "name": name, 'pid': os.getpid()}
    comm.send(msg, dest=0, tag=MPI_TAGS.READY)
    myid = 'worker {rank}/{name}/{pid}'.format(**msg)

    # wait for jobs
    while True:
        job_pickle = comm.recv(source=0, tag=MPI.ANY_TAG, status=status)
        tag = status.Get_tag()

        if tag == MPI_TAGS.START:
            jobid, job = loads(job_pickle)
            func, args, kwargs = job
            print("{myid}: runnin {func}".format(myid=myid, func=func))
            # Do the work here
            res = func(*args, **kwargs)
            print("{myid}: finished {func}".format(myid=myid, func=func))
            res_pickle = dumps(res)
            comm.send(res_pickle, dest=0, tag=MPI_TAGS.DONE)
            print("{myid}: sending result".format(myid=myid))
        elif tag == MPI_TAGS.EXIT:
            print("{myid}: exit".format(myid=myid))
            break


class MPIExecutor(object):
    """Executes jobs in local subprocesses using concurrent.futures
    """

    _jobid_counter = itertools.count()
    available_workers = []

    def __init__(self):
        self.comm = MPI.COMM_WORLD   # get MPI communicator object
        self.size = self.comm.size        # total number of processes
        self.rank = self.comm.rank        # rank of this process
        self.status = MPI.Status()   # get MPI status object
        # list of workers + their status
        self.workers = {}

        self.num_workers = self.size - 1
        if self.num_workers < 1:
            raise Exception('Not enough MPI workers')
        print("Master starting with %d workers" % self.num_workers)
        # collect worker READY messages
        # TODO choose minimum number and timeout
        while len(self.workers) < self.num_workers:
            data = self.comm.recv(source=MPI.ANY_SOURCE, tag=MPI_TAGS.READY, status=self.status)
            source = self.status.Get_source()
            self.workers[source] = {'online': True, 'ready_message': data}
            logger.info('worker {} in, {}/{}'.format(source, len(self.workers), self.num_workers))
            self.available_workers.append(source)
        self.workers_cycle = itertools.cycle(self.workers.keys())

    def submit(self, func, *args, **kwargs):
        """Submit a function: func(*args, **kwargs) and return a FutureJob.
        """
        # worker = self.available_workers.pop()
        # random choice workers
        # worker = random.choice(self.available_workers)
        # Round-Robin
        worker = next(self.workers_cycle)
        job = (func, args, kwargs)
        # jobid = hash(job)
        jobid = next(self._jobid_counter)
        job_pickle = dumps((jobid, job))
        self.comm.send(job_pickle, dest=worker, tag=MPI_TAGS.START)
        # job = {'worker': worker, 'jobid': jobid}
        return FutureMPIJob(jobid, job, worker, self)

    def shutdown(self):
        """Shut down the workers
        """
        logger.debug('MPI executor shutdown')
        for source, worker in self.workers.items():
            if worker['online']:
                logger.debug('Shut down MPI worker {}'.format(source))
                self.comm.send(None, dest=source, tag=MPI_TAGS.EXIT)
                worker['online'] = False

    def __del__(self):
        self.shutdown()
        if hasattr(super(type(self)), '__del__'):
            super(type(self), self).__del__()


class LocalExecutor(object):
    """Executes jobs in the current process
    """

    def __init__(self):
        super(LocalExecutor, self).__init__()

    def submit(self, func, *args, **kwargs):
        job = LocalFutureJob(func, *args, **kwargs)
        return job


class IpyparallelExecutor(object):
    """Executes jobs using ipyparallel

    Args:
        profiles (Optional[iterable]): list of ipyparallel profile names
        profile_dirs (Optional[iterable]): list of ipyparallel profile directories
        display_outputs (Optional[bool]): display stdout/err from actors [False]
        timeout (Optional): timeout in secs for waiting for ipyparallel cluster [60]
        min_engines (Optional[int]): minimum number of engines [1]
        client_kwargs: passed to ipyparallel.Client( **kwargs)
    """

    def __init__(self,
                 profiles=(),
                 profile_dirs=(),
                 display_outputs=False,
                 timeout=60,
                 min_engines=1,
                 client_kwargs=None):

        self.process_pool = []
        # actor: job
        frame = inspect.currentframe()
        args, varargs, keywords, values = inspect.getargvalues(frame)
        self._init_args = [values[k] for k in args[1:]]
        if keywords:
            self._init_kwargs = {k: values[k] for k in keywords}
        else:
            self._init_kwargs = {}
        self.display_outputs = display_outputs
        # get individual clients
        self._ipy_rc = []
        self._current_cli = 0
        self._ipy_lv = []
        self._ipy_dv = []
        # decide whether to use profiles or profile_dirs
        if profiles:
            cli_arg = 'profile'
            cli_args = profiles
        else:
            cli_arg = 'profile_dir'
            cli_args = profile_dirs
        if not cli_args:
            # raise ValueError('Either profiles or profile_dirs must be specified')
            cli_arg = 'profile'
            cli_args = ('default', )
        if isinstance(cli_args, six.string_types):
            cli_args = (cli_args, )

        for value in cli_args:
            # init ipyparallel clients
            kwargs = {cli_arg: value}
            if client_kwargs:
                kwargs.update(client_kwargs)
            # TODO min_engines divide by len(profile_dirs)
            self._ipy_rc.append(self.init_cluster(min_engines, timeout, **
                                                  kwargs))
            self._ipy_dv.append(self._ipy_rc[-1][:])
            self._ipy_lv.append(self._ipy_rc[-1].load_balanced_view())

        self.running_actors = {}
        self.execution_queue = deque()
        self.wait_queue = []

    @staticmethod
    def init_cluster(min_engines, timeout, *args, **kwargs):
        '''Get a connection (view) to an IPython cluster

        Args:
            *args: passed to ipyparallel.Client(*args, **kwargs)
            **kwargs: passed to ipyparallel.Client(*args, **kwargs)
        '''

        maxtime = time.time() + timeout

        while True:
            try:
                cli = Client(*args, **kwargs)
            except Exception as e:
                if time.time() > maxtime:
                    # raise the original exception from ipyparallel
                    raise e
                else:
                    # sleep and try again to get the client
                    print("Waiting for ipyparallel cluster Client(*{}, **{})".format(args, kwargs))
                    print(e)
                    time.sleep(timeout * 0.1)
                    continue
            if len(cli.ids) >= min_engines:
                # we have enough clients
                break
            else:
                # free the client
                cli.close()
            if time.time() > maxtime:
                raise Exception('Not enough ipyparallel clients')
            # try ~10 times
            print("Found {}/{} clients, need more ...".format(len(cli.ids), min_engines))
            time.sleep(timeout * 0.1)

        return cli

    def _rotate_client(self):
        # TODO pick the first (most) empty one
        current = self._current_cli
        self._current_cli = (self._current_cli + 1) % len(self._ipy_rc)
        return current

    def submit(self, func, *args, **kwargs):
        """Submit a function: func(*args, **kwargs) and return a FutureJob.
        """

        # print("Run actor {}".format(actor))
        lv = self._ipy_lv[self._rotate_client()]
        job = lv.apply_async(func, *args, **kwargs)
        return FutureIpyJob(job)


class FutureJob(object):
    """An asynchronous job with Future-like API
    """

    def __init__(self, future):
        self._future = future

    def __getattr__(self, item):
        return getattr(self._future, item)

    def display_outpus(self):
        # TODO how to display outputs?
        pass


class LocalFutureJob(object):
    """Local (system) job with FutureJob API
    """

    def __init__(self, func, *args, **kwargs):
        self.started = datetime.datetime.now()
        self._result = func(*args, **kwargs)

    def done(self):
        return True

    def result(self, timeout=None):
        return self._result

    def display_outputs(self):
        pass


class FutureIpyJob(object):
    """Wraps asynchronous results of different kinds into a future-like object
    """

    def __init__(self, job):
        self._job = job

    def done(self):
        return self._job.ready()

    def result(self, timeout=None):
        return self._job.get()

    def display_outputs(self):
        return self._job.display_outputs()


class FutureMPIJob(object):
    """Wraps asynchronous results of different kinds into a future-like object
    """

    _submitted = {}
    _received = {}

    def __init__(self, jobid, job, worker, executor):
        func, args, kwargs = job
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.worker = worker
        self.started = datetime.datetime.now()
        self.completed = None
        self.engine_id = worker
        self.error = None
        self.executor = executor
        self.request = MPI.COMM_WORLD.irecv(source=worker, tag=MPI_TAGS.DONE)
        self._done = False
        self._result = None

    def done(self):
        if not self._done:
            flag, rmess = self.request.test()
            self._done = flag
            self._result = rmess
        if self._done:
            self.completed = datetime.datetime.now()
            # self.executor.available_workers.append(self.worker)
        return self._done

    def result(self, timeout=None):
        if timeout is not None:
            raise NotImplementedError('finite timeout not implemented yet')
        if self.done():
            res_pickle = self._result
        else:
            res_pickle = self.request.wait()
            self._done = True
            self._result = res_pickle
            self.completed = datetime.datetime.now()
            # self.executor.available_workers.append(self.worker)
        res = loads(res_pickle)
        return res

    def display_outputs(self):
        return 'Sorry, not implemented for MPI :-('


class _IPySystemJob(object):
    """
    System (local run) IPython parallel like job
    """

    def __init__(self, actor, *args, **kwargs):
        self.actor = actor
        self.args = args
        self.kwargs = kwargs
        self.started = datetime.datetime.now()
        self._res = self.actor.run(*self.args, **self.kwargs)
        self.completed = datetime.datetime.now()
        self.engine_id = None
        self.error = None

    def ready(self):
        return True

    def get(self):
        return self._res

    def display_outputs(self):
        pass


class IPyClusterScheduler(_ActorRunner):
    """
    Scheduler using ipyparallel cluster.

    Args:
        display_outputs (Optional[bool]): display stdout/err from actors [False]
        timeout (Optional): timeout in secs for waiting for ipyparallel cluster [60]
        min_engines (Optional[int]): minimum number of engines [1]
        *args: passed to self.init_cluster(*args, **kwargs)
        **kwargs: passed to self.init_cluster(*args, **kwargs)
    """

    def __init__(self, *args, **kwargs):
        self.process_pool = []
        # actor: job
        self._init_args = args
        self._init_kwargs = kwargs
        self.display_outputs = kwargs.pop('display_outputs', False)
        self.timeout = kwargs.pop('timeout', 60)
        self.min_engines = kwargs.pop('min_engines', 1)
        self.running_actors = {}
        self.execution_queue = deque()
        self.wait_queue = []
        self._ipy_rc = self.init_cluster(self.min_engines, self.timeout, *args,
                                         **kwargs)
        self._ipy_dv = self._ipy_rc[:]
        self._ipy_lv = self._ipy_rc.load_balanced_view()

    def copy(self):
        return self.__class__(*self._init_args, **self._init_kwargs)

    def put_value(self, in_port, value):
        self.execution_queue.appendleft((in_port, value))

    @staticmethod
    def init_cluster(min_engines, timeout, *args, **kwargs):
        '''Get a connection (view) to an IPython cluster

        Args:
            *args: passed to ipyparallel.Client(*args, **kwargs)
            **kwargs: passed to ipyparallel.Client(*args, **kwargs)
        '''

        maxtime = time.time() + timeout

        while True:
            try:
                cli = Client(*args, **kwargs)
            except Exception as e:
                if time.time() > maxtime:
                    # raise the original exception from ipyparallel
                    raise e
                else:
                    # sleep and try again to get the client
                    time.sleep(timeout * 0.1)
                    continue
            if len(cli.ids) >= min_engines:
                # we have enough clients
                break
            else:
                # free the client
                cli.close()
            if time.time() > maxtime:
                raise Exception('Not enough ipyparallel clients')
            # try ~10 times
            time.sleep(timeout * 0.1)

        return cli

    def execute(self):

        self.reset()

        while self.execution_queue or self.running_actors or self.wait_queue:
            self.nothing = True

            self._try_empty_execution_queue()
            self._try_empty_wait_queue()
            self._try_empty_ready_jobs()

            # TODO fix against spamming engines - PROBABLY NOT THE BEST WAY!
            if self.nothing:
                # increase sleep time if nothing is happening
                self.last_sleep = min(1, max(0.001, self.last_sleep * 2))
                logger.debug('scheduler sleeps for {} ms'.format(
                    self.last_sleep * 1e3))
                time.sleep(self.last_sleep)
            else:
                self.last_sleep = 0

    def _try_empty_ready_jobs(self):
        pending = {}  # temporary container
        for actor, job_description in self.running_actors.items():

            job = job_description['job']

            if 'started' not in job_description and job_description[
                    'job'].started:
                # log the started time
                job_description['started'] = job_description['job'].started
                job_description['engine'] = job_description['job'].engine_id
                logger.debug(
                    "started {started} actor {actor}({args}, {kwargs})"
                    " on engine {engine}".format(actor=actor.name,
                                                 **job_description))
            if job.ready():
                self.nothing = False
                if 'started' not in job_description and job_description[
                        'job'].started:
                    # log the started time
                    job_description['started'] = job_description['job'].started
                    job_description['engine'] = job_description[
                        'job'].engine_id
                    logger.debug(
                        "started {started} actor {actor}({args}, {kwargs})"
                        " on engine {engine} at {started}".format(
                            actor=actor.name,
                            **job_description))

                job_description['completed'] = job_description['job'].completed
                logger.debug('completed actor {actor} at {completed}'.format(
                    actor=actor.name,
                    **job_description))

                # process result
                # raise RemoteError in case of failure
                try:
                    result = job.get()
                except RemoteError:
                    logger.error('actor {} failed\n{}'.format(
                        actor.name, job_description['job'].error))
                    raise
                if self.display_outputs:
                    job.display_outputs()
                if result:
                    # empty results don't need any processing
                    out_names = actor.outports.keys()
                    if not hasattr(result, 'items'):
                        raise ValueError(
                            'The execute method must return '
                            'a dict-like object with items method')
                    for name, value in result.items():
                        if name in out_names:
                            outport = actor.outports[name]
                            outport.put(value)
                            self.on_outport_put_value(outport)
                        else:
                            raise ValueError("{} not in output ports".format(
                                name))

            else:
                pending[actor] = job_description
        self.running_actors = pending

    def _try_empty_wait_queue(self):
        pending = []  # temporary container
        for actor in self.wait_queue:
            # run actors only if not already running
            if actor not in self.running_actors:
                self.nothing = False
                # TODO can we iterate and remove at the same time?
                self.running_actors[actor] = self.run_actor(actor)
            else:
                pending.append(actor)
        self.wait_queue = pending

    def _try_empty_execution_queue(self):
        while self.execution_queue:
            in_port, value = self.execution_queue.pop()
            should_run = in_port.put(value)
            if should_run:
                self.nothing = False
                # waiting to be run
                self.wait_queue.append(in_port.owner)
                # self.running_actors((in_port.owner, self.run_actor(in_port.owner)))

    def run_actor(self, actor):
        # print("Run actor {}".format(actor))
        actor.scheduler = self
        args, kwargs = actor.get_run_args()
        # system actors must be run within this process
        res = dict(args=args, kwargs=kwargs)
        if actor.system_actor:
            res['job'] = _IPySystemJob(actor, *args, **kwargs)
        else:
            res['job'] = self._ipy_lv.apply_async(actor.run, *args, **kwargs)

        logger.debug('submitted actor {}({}, {})'.format(actor.name, args,
                                                         kwargs))

        return res


class MultiIpyClusterScheduler(IPyClusterScheduler):
    """Scheduler using multiple ipyparallel clusters.

    Can use more than 1 ipyparallel cluster in order to avoid limitations for the number of engines.
    Either profiles or profile_dirs must be specified.

    Args:
        profiles (Optional[iterable]): list of ipyparallel profile names
        profile_dirs (Optional[iterable]): list of ipyparallel profile directories
        display_outputs (Optional[bool]): display stdout/err from actors [False]
        timeout (Optional): timeout in secs for waiting for ipyparallel cluster [60]
        min_engines (Optional[int]): minimum number of engines [1]
        client_kwargs: passed to ipyparallel.Client( **kwargs)
    """

    def __init__(self,
                 profiles=(),
                 profile_dirs=(),
                 display_outputs=False,
                 timeout=60,
                 min_engines=1,
                 client_kwargs=None):
        self.process_pool = []
        # actor: job
        frame = inspect.currentframe()
        args, varargs, keywords, values = inspect.getargvalues(frame)
        self._init_args = [values[k] for k in args[1:]]
        if keywords:
            self._init_kwargs = {k: values[k] for k in keywords}
        else:
            self._init_kwargs = {}
        self.display_outputs = display_outputs
        self.timeout = timeout
        self.min_engines = min_engines
        # get individual clients
        self._ipy_rc = []
        self._current_cli = 0
        self._ipy_lv = []
        self._ipy_dv = []
        # decide whether to use profiles or profile_dirs
        if profiles:
            cli_arg = 'profile'
            cli_args = profiles
        else:
            cli_arg = 'profile_dir'
            cli_args = profile_dirs
        if not cli_args:
            raise ValueError(
                'Either profiles or profile_dirs must be specified')

        for value in cli_args:
            # init ipyparallel clients
            kwargs = {cli_arg: value}
            if client_kwargs:
                kwargs.update(client_kwargs)
            self._ipy_rc.append(self.init_cluster(min_engines, timeout, **
                                                  kwargs))
            self._ipy_dv.append(self._ipy_rc[-1][:])
            self._ipy_lv.append(self._ipy_rc[-1].load_balanced_view())

        self.running_actors = {}
        self.execution_queue = deque()
        self.wait_queue = []

    def _rotate_client(self):
        current = self._current_cli
        self._current_cli += 1
        if self._current_cli == len(self._ipy_rc):
            self._current_cli = 0
        return current

    def run_actor(self, actor):
        # print("Run actor {}".format(actor))
        actor.scheduler = self
        args, kwargs = actor.get_run_args()
        # system actors must be run within this process
        res = dict(args=args, kwargs=kwargs)
        lv = self._ipy_lv[self._rotate_client()]
        if actor.system_actor:
            res['job'] = _IPySystemJob(actor, *args, **kwargs)
        else:
            res['job'] = lv.apply_async(actor.run, *args, **kwargs)

        logger.debug('submitted actor {}({}, {})'.format(actor.name, args,
                                                         kwargs))

        return res


class FuturesScheduler(_ActorRunner):
    """Scheduler using PEP 3148 futures

    Args:
        executor (str): executor type: multiprocessing, distributed, ipyparallel, mpi
        display_outputs (Optional[bool]): display stdout/err from actors [False]
        timeout (Optional): timeout in secs for waiting for ipyparallel cluster [60]
        min_engines (Optional[int]): minimum number of engines [1]
        client_kwargs: passed to ipyparallel.Client( **kwargs)
    """

    def __init__(self,
                 executor,
                 display_outputs=False,
                 min_engines=1,
                 timeout=60,
                 executor_kwargs=None,
                 copy_from=None):

        if executor_kwargs is None:
            executor_kwargs = {}

        # for .copy()
        # frame = inspect.currentframe()
        # args, varargs, keywords, values = inspect.getargvalues(frame)
        # self._init_args = [values[k] for k in args[1:]]
        # if keywords:
        #     self._init_kwargs = {k: values[k] for k in keywords if k not in ('copy_from', )}
        # else:
        #     self._init_kwargs = {}
        self._init_args = (executor, )
        self._init_kwargs = {'display_outputs': display_outputs,
                             'min_engines': min_engines,
                             'timeout': timeout,
                             'executor_kwargs': executor_kwargs}
        self.display_outputs = display_outputs

        if copy_from is None:
            if executor == 'multiprocessing':
                self.executor = MultiprocessingExecutor(processes=min_engines)
            elif executor == 'distributed':
                self.executor = DistributedExecutor(uris=executor_kwargs['uris'],
                                                    min_engines=min_engines,
                                                    timeout=timeout)
            elif executor == 'ipyparallel':
                kwargs = dict(min_engines=min_engines,
                              timeout=timeout,
                              display_outputs=display_outputs)
                kwargs.update(executor_kwargs)
                self.executor = IpyparallelExecutor(**kwargs)
            elif executor == 'mpi':
                self.executor = MPIExecutor()
            # elif executor == 'scoop':
            #     self.executor = ScoopExecutor()
            else:
                raise ValueError('Executor {} unknown'.format(executor))
            self.system_executor = LocalExecutor()
        else:
            # executors must be shared across copies to avoid their initialization
            self.executor = copy_from.executor
            self.system_executor = copy_from.system_executor

        self.reset()

    def reset(self):
        self.process_pool = []
        self.running_actors = {}
        self.execution_queue = deque()
        self.wait_queue = []
        self.last_sleep = 0

    def run_actor(self, actor):
        # print("Run actor {}".format(actor))
        actor.scheduler = self
        args, kwargs = actor.get_run_args()
        # system actors must be run within this process
        res = dict(args=args, kwargs=kwargs)
        if actor.system_actor:
            res['job'] = self.system_executor.submit(actor.run, *args, **
                                                     kwargs)
        else:
            res['job'] = self.executor.submit(actor.run, *args, **kwargs)

        logger.debug('submitted actor {}({}, {})'.format(actor.name, args,
                                                         kwargs))

        return res

    def copy(self):
        return self.__class__(*self._init_args, copy_from=self, **self._init_kwargs)

    def put_value(self, in_port, value):
        self.execution_queue.appendleft((in_port, value))

    def execute(self):
        while self.execution_queue or self.running_actors or self.wait_queue:
            self.nothing = True

            self._try_empty_execution_queue()
            self._try_empty_wait_queue()
            self._try_empty_ready_jobs()

            # TODO fix against spamming engines - PROBABLY NOT THE BEST WAY!
            if self.nothing:
                # increase sleep time if nothing is happening
                self.last_sleep = min(1, max(0.001, self.last_sleep * 2))
                logger.debug('scheduler sleeps for {} ms'.format(
                    self.last_sleep * 1e3))
                time.sleep(self.last_sleep)
            else:
                self.last_sleep = 0

                # TODO could we use callbacks?

    def _try_empty_ready_jobs(self):
        pending = {}  # temporary container
        for actor, job_description in self.running_actors.items():

            job = job_description['job']

            if job.done():
                self.nothing = False
                # process result
                # raise RemoteError in case of failure
                try:
                    result = job.result()
                except Exception:
                    logger.error('actor {} failed\n{}'.format(
                        actor.name, traceback.format_exc()))
                    self.reset()
                    raise
                if self.display_outputs:
                    job.display_outputs()
                if result:
                    # empty results don't need any processing
                    out_names = actor.outports.keys()
                    if not hasattr(result, 'items'):
                        raise ValueError(
                            'The execute method must return '
                            'a dict-like object with items method')
                    for name, value in result.items():
                        if name in out_names:
                            outport = actor.outports[name]
                            outport.put(value)
                            self.on_outport_put_value(outport)
                        else:
                            raise ValueError("{} not in output ports".format(
                                name))

            else:
                pending[actor] = job_description
        self.running_actors = pending

    def _try_empty_wait_queue(self):
        pending = []  # temporary container
        for actor in self.wait_queue:
            # run actors only if not already running
            if actor not in self.running_actors:
                self.nothing = False
                # TODO can we iterate and remove at the same time?
                self.running_actors[actor] = self.run_actor(actor)
            else:
                pending.append(actor)
        self.wait_queue = pending

    def _try_empty_execution_queue(self):
        while self.execution_queue:
            in_port, value = self.execution_queue.pop()
            should_run = in_port.put(value)
            if should_run:
                self.nothing = False
                # waiting to be run
                self.wait_queue.append(in_port.owner)
                # self.running_actors((in_port.owner, self.run_actor(in_port.owner)))

    def shutdown(self):
        logger.info('Scheduler is shutting down')
        if hasattr(self.executor, 'shutdown'):
            self.executor.shutdown()

    def __del__(self):
        self.shutdown()
        if hasattr(super(type(self)), '__del__'):
            super(type(self), self).__del__()


class ThreadedSchedulerWorker(threading.Thread, _ActorRunner):
    """Thread object that executes run after run of the ThreadedScheduler actors.
    """

    def __init__(self, scheduler, inner_id):
        threading.Thread.__init__(self)
        self.scheduler = scheduler
        self.executing = False
        self.finished = False
        self.inner_id = inner_id
        self.state_mutex = threading.RLock()

    def run(self):
        from time import sleep
        while not self.finished:
            pv = self.scheduler.pop_idle_task()
            if pv:
                port, value = pv
                should_run = port.put(value)  # Change to if
                if should_run:
                    # print(self.inner_id, port.owner.name, value)
                    self.run_actor(port.owner)
                else:
                    pass
                    # print(self.inner_id, "Won't run", )
                self.scheduler.on_actor_finished(port.owner)
            else:
                pass
                # print(self.inner_id, "Nothing to do")
                sleep(0.02)
                # print(self.inner_id, "End")

    def put_value(self, in_port, value):
        # print(self.inner_id, " worker put ", value)
        self.scheduler.put_value(in_port, value)

    def finish(self):
        """Finish after the current running job is done."""
        with self.state_mutex:
            self.finished = True


class ThreadedScheduler(object):
    def __init__(self, max_threads=2):
        self.max_threads = max_threads
        self.threads = []
        self.execution_queue = deque()
        self.running_actors = []
        self.state_mutex = threading.RLock()

    def copy(self):
        return self.__class__(max_threads=self.max_threads)

    def pop_idle_task(self):
        with self.state_mutex:
            for port, value in self.execution_queue:
                if port.owner not in self.running_actors:
                    # Removes first occurrence - it's probably safe
                    # print("Removing", value)
                    self.execution_queue.remove((port, value))
                    self.running_actors.append(port.owner)
                    return port, value
            else:
                return None

    def put_value(self, in_port, value):
        with self.state_mutex:  # Probably not necessary
            # print("put ", value, ", in queue: ", len(self.queue))
            self.execution_queue.append((in_port, value))

    def is_running(self):
        with self.state_mutex:
            return bool(self.running_actors or self.execution_queue)

    def on_actor_finished(self, actor):
        with self.state_mutex:
            self.running_actors.remove(actor)
            if not self.is_running():
                self.finish_all_threads()

    def execute(self):
        with self.state_mutex:  # Probably not necessary
            for i in range(self.max_threads):
                thread = ThreadedSchedulerWorker(self, i)
                self.threads.append(thread)
                thread.start()
                # print("Thread started", thread.ident)
        for thread in self.threads:
            # print("Join thread")
            thread.join()

    def finish_all_threads(self):
        # print("Everything finished. Waiting for threads to end.")
        for thread in self.threads:
            thread.finish()

    def shutdown(self):
        pass



from __future__ import absolute_import, division, print_function, unicode_literals
import tempfile
import os


def build_graph(actor):
    """Create graph with all actors + ports as nodes.
    
    It walks over all connections.
    
    Prerequisities:
    * graphviz package
    * dot executable (graphviz)
    
    Empty input ports are yellow.
    Empty output ports are red.
    """
    import graphviz
    actors = []
    edges = []
    ports = []
    graph = graphviz.Digraph()

    def _get_name(obj):
        return str(hash(obj))

    def _add_node(actor):
        attrs = {"fontsize": "12"}
        graph.node(_get_name(actor), label=actor.name, shape="box", **attrs)
        actors.append(actor)

    def _walk_node(actor):
        name = _get_name(actor)
        if not actor in actors:
            _add_node(actor)
        for port in actor.outports:
            if port not in ports:
                attrs = {}
                if not port.connections:
                    attrs["style"] = "filled"
                    attrs["color"] = "#ff0000"
                graph.node(_get_name(port), label=port.name, **attrs)
                graph.edge(name, _get_name(port))
                ports.append(port)
            for other in port.connections:
                if (port, other) not in edges:
                    edges.append((port, other))
                    _walk_node(other.owner)
                    graph.edge(_get_name(port), _get_name(other))
        for port in actor.inports:
            if port not in ports:
                attrs = {}
                if not port.connections:
                    attrs["style"] = "filled"
                    attrs["color"] = "#ffff00"
                graph.node(_get_name(port), label=port.name, **attrs)
                graph.edge(_get_name(port), name, )
                ports.append(port)
            for other in port.connections:
                if (other, port) not in edges:
                    _walk_node(other.owner)

    _walk_node(actor)
    return graph


def ipy_show(actor):
    """Display graph in IPython.

    """
    graph = build_graph(actor)
    graph.format = "png"

    from IPython.display import Image
    d = tempfile.mkdtemp()
    basename = os.path.join(d, "image")
    graph.render(basename)

    return Image(os.path.join(d, 'image.png'))

    # TODO: clean-up tempdir



#!/usr/bin/env python
from mpi4py import MPI
import numpy
import os
import cloudpickle


comm = MPI.Comm.Get_parent()
size = comm.Get_size()
rank = comm.Get_rank()


N = numpy.array(0, dtype='i')

spam_p = None
comm.bcast(spam_p, root=0)
print(spam_p)
spam = cloudpickle.loads(spam_p)

comm.Bcast([N, MPI.INT], root=0)
print('worker rank {}, size {}, N {}, pid {}'.format(rank, size, N, os.getpid()))
h = 1.0 / N
s = 0.0
for i in range(rank, N, size):
    x = h * (i + 0.5)
    s += 4.0 / (1.0 + x**2)
    # s += spam(x)
PI = numpy.array(s * h, dtype='d')
comm.Reduce([PI, MPI.DOUBLE], None,
            op=MPI.SUM, root=0)

comm.Disconnect()



from __future__ import absolute_import, division, print_function
from ..components import Actor
import inspect
import itertools
import six

__all__ = ['FuncActor', 'Switch', 'ShellRunner', 'Sink', 'DictionaryMerge', 'LoopWhile']


class FuncActor(Actor):
    """Actor defined simply by a function

    Args:
        func (callable): The function to be called on actor execution
        args (list): Fixed function positional arguments
        kwargs(dict): Fixed function keyword arguments
            args and kwargs behave as functools.partial args and keywords
        outports: output port name(s)
        inports: input port name(s)
        name (str): actor name
    """

    def __init__(self, func, args=(), kwargs={}, outports=None, inports=None, name=None):
        if not name:
            name = func.__name__
        super(FuncActor, self).__init__(name=name)
        try:
            # try to derive ports from function signature
            if six.PY2:
                # Python 2 does not have signatures
                fargs = inspect.getargspec(func)
                if inports is None:
                    inports = (par
                               for par in itertools.islice(fargs.args, len(args), None)
                               if par not in kwargs)
            else:
                sig = inspect.signature(func)
                return_annotation = sig.return_annotation
                # derive ports from func signature
                if inports is None:
                    # filter out args (first len(args) arguments and kwargs)
                    inports = (
                        par.name
                        for par in itertools.islice(sig.parameters.values(), len(args), None)
                        if par.name not in kwargs)
                if outports is None and return_annotation is not inspect.Signature.empty:
                    # if func has a return annotation, use it for outports names
                    outports = return_annotation
        except (ValueError, TypeError):
            # e.g. numpy has no support for inspect.signature
            # --> using manual inports
            if inports is None:
                inports = ('inp', )
            elif isinstance(inports, six.string_types):
                inports = (inports, )
        # save func as attribute
        self.func = func
        self._func_args = args
        self._func_kwargs = kwargs
        # setup inports
        for name in inports:
            self.inports.append(name)
        # setup outports
        if outports is None:
            outports = ('out', )
        elif isinstance(outports, six.string_types):
            outports = (outports, )
        for name in outports:
            self.outports.append(name)

    def get_run_args(self):
        args = tuple(port.pop() for port in self.inports)
        kwargs = {'runfunc': self.func,
                  'func_args': self._func_args,
                  'func_kwargs': self._func_kwargs,
                  'outports': tuple(port.name for port in self.outports)}
        # kwargs['connected_ports'] = list((name for name, port in self.outports.items()
        #                                   if port.isconnected()))

        return args, kwargs

    @staticmethod
    def run(*args, **kwargs):
        args = kwargs['func_args'] + args
        func_res = kwargs['runfunc'](*args, **kwargs['func_kwargs'])
        outports = kwargs['outports']

        if len(outports) == 1:
            func_res = (func_res, )
        # iterate over ports and return values
        res = {name: value for name, value in zip(outports, func_res)}
        return res

    def __call__(self, *args, **kwargs):
        args = self._func_args + args
        kwargs.update(self._func_kwargs)
        return self.func(*args, **kwargs)


class Switch(Actor):
    """Redirects to either 'true' or 'false' output based on the condition value
    """

    _system_actor = True

    def __init__(self, name=None, condition_func=None):
        super(Switch, self).__init__(name=name)
        self._in_condition = False
        self.inports.append('inp')
        self.outports.append('true')
        self.outports.append('false')
        self.outports.append('condition_in')
        self.inports.append('condition_out')
        if condition_func is None:
            self._condition_func = None
        elif callable(condition_func):
            self._condition_func = condition_func
        else:
            raise Exception('condition_func must be a callable object')

    def get_run_args(self):
        # everything is done inside run
        return (), {}

    def is_condition_actor(self):
        """Returns True if condition actor is connected
        """
        if (self.inports['condition_out'].isconnected() and
                self.outports['condition_in'].isconnected()):
            return True
        elif (self.inports['condition_out'].isconnected() or
              self.outports['condition_in'].isconnected()):
            raise Exception('Both condition_in and out must be connected')
        return False

    def run(self, *args, **kwargs):
        res = {}
        condition_out = None
        if not self._in_condition:
            # input on init port
            value = self.inports['inp'].pop()

        elif not self.inports['condition_out'].isempty():
            # we receive the condition actor output
            # the value was stored
            value = self._last_value
            condition_out = self.inports['condition_out'].pop()
            self._in_condition = False
        else:
            raise Exception('Enexpected error')

        if condition_out is None:
            # we have to evaluate the condition
            if self.is_condition_actor():
                self._last_value = value
                self._in_condition = True
                res['condition_in'] = value
                # we have to return here to execute the condition actor
                return res
            else:
                condition_out = self._condition_func(value)

        if condition_out:
            res['true'] = value
        else:
            res['false'] = value
        return res

    def can_run(self):
        if self._in_condition:
            # waiting for the condition actor
            res = not self.inports['condition_out'].isempty()
        else:
            res = not self.inports['inp'].isempty()
        return res


class ShellRunner(Actor):
    """An actor executing external command.

    Basically, it calls subprocess.call(base_command + inp) or
    subprocess.call(base_command.format(inp)) in case format_inp is True.

    Args:
        base_command: the command to be run, may be a template
        binary: input/output in binary mode
        shell: shell parameter in subprocess.call
               if it's a string then it indicates the executable parameter in subprocess.call
        format_inp: 'args' triggers base_command.format(*inp),
                    'kwargs' triggers base_command.format(**inp)
        single_out: join outputs into a single dict
        debug_print: print debug info
        print_output: print standard output and standard error
    """

    def __init__(self,
                 base_command,
                 name=None,
                 binary=False,
                 shell=False,
                 format_inp=False,
                 single_out=False,
                 print_output=False,
                 debug_print=False):
        super(ShellRunner, self).__init__(name=name)

        if isinstance(base_command, six.string_types):
            self.base_command = (base_command, )
        else:
            self.base_command = base_command

        self.binary = binary
        self.shell = shell
        self.format_inp = format_inp
        self.inports.append('inp')
        self.single_out = single_out
        self.debug_print = debug_print
        self.print_output = print_output
        if single_out:
            self.outports.append('out')
        else:
            self.outports.append('stdout')
            self.outports.append('stderr')
            self.outports.append('ret')

    def get_run_args(self):
        vals = self.inports['inp'].pop()
        if self.format_inp == 'args':
            args = (self.base_command[0].format(*vals), )
        elif self.format_inp == 'kwargs':
            args = (self.base_command[0].format(**vals), )
        elif self.format_inp == 'trigger':
            args = self.base_command
        elif isinstance(vals, six.string_types):
            args = self.base_command + (vals, )
        else:
            args = self.base_command + vals
        kwargs = {
            'shell': self.shell,
            'binary': self.binary,
            'single_out': self.single_out,
            'debug_print': self.debug_print,
            'print_output': self.print_output,
        }
        return args, kwargs

    @staticmethod
    def run(*args, **kwargs):
        import subprocess
        import tempfile
        import sys

        if kwargs['debug_print']:
            print('run command:\n{}'.format(' '.join(args)))

        if kwargs['binary']:
            mode = "w+b"
        else:
            mode = "w+t"

        with tempfile.TemporaryFile(mode=mode) as fout, tempfile.TemporaryFile(
                mode=mode) as ferr:
            if kwargs['shell']:

                executable = kwargs['shell']
                if not isinstance(kwargs['shell'], six.string_types):
                    executable = None

                result = subprocess.call(' '.join(args),
                                         stdout=fout,
                                         stderr=ferr,
                                         executable=executable,
                                         shell=kwargs['shell'])
            else:
                result = subprocess.call(args,
                                         stdout=fout,
                                         stderr=ferr,
                                         shell=kwargs['shell'])
            fout.seek(0)
            ferr.seek(0)
            cout = fout.read()
            cerr = ferr.read()
        res = {'ret': result, 'stdout': cout, 'stderr': cerr}
        if kwargs['debug_print']:
            print('result:\n{}'.format(res))
        if kwargs['print_output']:
            sys.stdout.write(cout)
            sys.stderr.write(cerr)
        if kwargs['single_out']:
            res = {'out': res}
        return res


class Sink(Actor):
    """Dumps everything
    """

    _system_actor = True

    def can_run(self):
        return True

    def get_run_args(self):
        for port in self.inports:
            port.pop()
        return (), {}

    @staticmethod
    def run(*args, **kwargs):
        pass


class DictionaryMerge(Actor):
    """This actor merges inputs from all its inports into one dictionary.

    The keys of the dictionary will be equal to inport names.
    """

    def __init__(self, name="packager", inport_names=("in"), outport_name="out"):
        super(DictionaryMerge, self).__init__(name=name)
        for in_name in inport_names:
            self.inports.append(in_name)
        self.outport_name = outport_name
        self.outports.append(outport_name)

    def get_run_args(self):
        return (), {
            "values": {port.name: port.pop()
                       for port in self.inports},
            "outport_name": self.outport_name
        }

    @staticmethod
    def run(*args, **kwargs):
        return {kwargs.get("outport_name"): kwargs.get("values")}


class LoopWhile(Actor):
    """A while loop actor"""

    _system_actor = True

    def __init__(self, name='LoopWhile', condition_func=None):
        super(LoopWhile, self).__init__(name=name)
        # flag for being inside a loop
        self._in_loop = False
        # flag for evaluating the condition
        self._in_condition = False
        # setup ports
        self.inports.append('init')
        self.inports.append('loop')
        self.outports.append('loop')
        self.outports.append('exit')
        # TODO: we could possibly create condition ports only if condition_func is None
        # --> would have to fix the logic in run
        self.outports.append('condition_in')
        self.inports.append('condition_out')
        if condition_func is None:
            self._condition_func = None
        elif callable(condition_func):
            self._condition_func = condition_func
        else:
            raise Exception('condition_func must be a callable object')

    def get_run_args(self):
        # everything is done inside run
        return (), {}

    def is_condition_actor(self):
        """Returns True if condition actor is connected
        """
        if (self.inports['condition_out'].isconnected() and
                self.outports['condition_in'].isconnected()):
            return True
        elif (self.inports['condition_out'].isconnected() or
              self.outports['condition_in'].isconnected()):
            raise Exception('Both condition_in and out must be connected')
        return False

    def run(self, *args, **kwargs):
        res = {}
        condition_out = None
        if not self._in_loop:
            # input on init port
            value = self.inports['init'].pop()
            self._in_loop = True
        elif not self.inports['loop'].isempty():
            # input value from the loop
            value = self.inports['loop'].pop()
        elif not self.inports['condition_out'].isempty():
            # we receive the condition actor output
            # the value was stored
            value = self._last_value
            condition_out = self.inports['condition_out'].pop()
        else:
            raise Exception('Enexpected error')

        if condition_out is None:
            # we have to evaluate the condition
            if self.is_condition_actor():
                self._last_value = value
                res['condition_in'] = value
                # we have to return here to execute the condition actor
                return res
            else:
                condition_out = self._condition_func(value)
        if condition_out:
            res['loop'] = value
        else:
            # this is the end - condition is False
            self._in_loop = False
            res['exit'] = value
        return res

    def can_run(self):
        if self._in_loop:
            # waiting for loop
            res = (not self.inports['loop'].isempty() or
                   not self.inports['condition_out'].isempty())
        else:
            res = not self.inports['init'].isempty()
        return res



from . import Actor


class Rand(Actor):

    def __init__(self, name=None):
        super(Rand, self).__init__(name=name)
        self.inports.append('n')
        self.outports.append('out')

    def get_args(self):
        args = (self.inports['n'].pop(), )
        kwargs = {}
        return args, kwargs

    @staticmethod
    def get_result(*args, **kwargs):

        from numpy.random import rand
        return {'out': rand(*args)}

    def run(self):
        args, kwargs = self.get_args()
        return self.get_result(*args, **kwargs)


class RandInt(Actor):

    def __init__(self, name=None):
        super(RandInt, self).__init__(name=name)
        self.inports.append('low')
        self.outports.append('out')

    def get_args(self):
        args = (1, self.inports['low'].pop(), )
        kwargs = {}
        return args, kwargs

    @staticmethod
    def get_result(*args, **kwargs):

        from numpy.random import randint
        return {'out': randint(*args)}

    def run(self):
        args, kwargs = self.get_args()
        return self.get_result(*args, **kwargs)



'''WOWP -- A WOrkfloW Framework in Python
'''
from __future__ import absolute_import, division, print_function, unicode_literals
from .__about__ import __version__, __release__

from .components import Component, Actor, Workflow
from . import actors

__all__ = "actors", "Component", "Actor", "Workflow"



from . import Actor
from julia import Julia, JuliaError
import six

class JuliaMethod(Actor):
    def __init__(self, method_name, package_name=None, inports=(), outports='result'):
        self.method_name = method_name
        self.package_name = package_name
        super(JuliaMethod, self).__init__(name=self._full_method_name)

        if isinstance(inports, six.string_types):
            inports = (inports,)
        for p in inports:
            self.inports.append(p)

        if isinstance(outports, six.string_types):
            outports = (outports,)
        for p in outports:
            self.outports.append(p)

    @property
    def _julia_method(self):
        if self.package_name:
            self._julia.eval("using %s" % self.package_name)
        return self._julia.eval(self._full_method_name)

    @property
    def _full_method_name(self):
        if self.package_name:
            return "%s.%s" % (self.package_name, self.method_name)
        else:
            return self.method_name

    def run(self):
        args = (port.pop() for port in self.inports)
        func_res = self.__call__(*args)

        if len(self.outports) == 1:
            func_res = (func_res,)

        res = {}
        for name, value in zip(self.outports.keys(), func_res):
            res[name] = value
        return res

    def __call__(self, *args, **kwargs):
        self._julia = Julia()
        return self._julia_method(*args)



"""
Tools to help with setup.py
Bootstrapped from https://github.com/jakevdp/mpld3.
Much of this is based on tools in the IPython project:
http://github.com/ipython/ipython
"""

import os


def get_version():
    """Get the version info from the package without importing it"""
    with open(os.path.join("wowp", "__about__.py"), "r") as init_file:
        exec(compile(init_file.read(), 'wowp/__about__.py', 'exec'), globals())
    try:
        return __version__
    except NameError:
        raise ValueError("version could not be located")



#!/usr/bin/env python
"""Demonstrate the task-pull paradigm for high-throughput computing
using mpi4py. Task pull is an efficient way to perform a large number of
independent tasks when there are more tasks than processors, especially
when the run times vary for each task.
This code is over-commented for instructional purposes.
This example was contributed by Craig Finch (cfinch@ieee.org).
Inspired by http://math.acadiau.ca/ACMMaC/Rmpi/index.html
"""

from mpi4py import MPI

def enum(*sequential, **named):
    """Handy way to fake an enumerated type in Python
    http://stackoverflow.com/questions/36932/how-can-i-represent-an-enum-in-python
    """
    enums = dict(zip(sequential, range(len(sequential))), **named)
    return type('Enum', (), enums)

# Define MPI message tags
tags = enum('READY', 'DONE', 'EXIT', 'START')

# Initializations and preliminaries
comm = MPI.COMM_WORLD   # get MPI communicator object
size = comm.size        # total number of processes
rank = comm.rank        # rank of this process
status = MPI.Status()   # get MPI status object

if rank == 0:
    # Master process executes code below
    tasks = range(2*size)
    task_index = 0
    num_workers = size - 1
    closed_workers = 0
    print("Master starting with %d workers" % num_workers)
    while closed_workers < num_workers:
        data = comm.recv(source=MPI.ANY_SOURCE, tag=MPI.ANY_TAG, status=status)
        source = status.Get_source()
        tag = status.Get_tag()
        if tag == tags.READY:
            # Worker is ready, so send it a task
            if task_index < len(tasks):
                comm.send(tasks[task_index], dest=source, tag=tags.START)
                # comm.send(tasks[task_index], dest=MPI.ANY_SOURCE, tag=tags.START)
                print("Sending task %d to worker %d" % (task_index, source))
                task_index += 1
            else:
                comm.send(None, dest=source, tag=tags.EXIT)
        elif tag == tags.DONE:
            results = data
            print("Got data from worker %d" % source)
        elif tag == tags.EXIT:
            print("Worker %d exited." % source)
            closed_workers += 1

    print("Master finishing")
else:
    # Worker processes execute code below
    name = MPI.Get_processor_name()
    print("I am a worker with rank %d on %s." % (rank, name))
    while True:
        comm.send(None, dest=0, tag=tags.READY)
        task = comm.recv(source=0, tag=MPI.ANY_TAG, status=status)
        tag = status.Get_tag()

        if tag == tags.START:
            # Do the work here
            result = task**2
            comm.send(result, dest=0, tag=tags.DONE)
        elif tag == tags.EXIT:
            break

    comm.send(None, dest=0, tag=tags.EXIT)



# -*- coding: utf-8 -*-
'''
Matlab actors

Requires MATLAB Engine for Python
(http://www.mathworks.com/help/matlab/matlab-engine-for-python.htm).
'''

from . import Actor
import stopit
import matlab.engine
from warnings import warn
import six


class EngineManager(object):
    """Matlab engine with a shared pool"""

    _engine_pool = []

    def __init__(self, engine_args=None, engine_kwargs=None):
        if engine_args is None:
            engine_args = ()
        if engine_kwargs is None:
            engine_kwargs = {}
        self.engine_args = engine_args
        self.engine_kwargs = engine_kwargs

    def _new_engine(self):
        return matlab.engine.start_matlab(*self.engine_args, **self.engine_kwargs)

    def pop(self):
        try:
            eng = self.__class__._engine_pool.pop()
        except IndexError:
            eng = self._new_engine()
        return eng

    def push(self, eng):
        self.__class__._engine_pool.append(eng)

    @classmethod
    def close_all(cls, timeout=10):
        while cls._engine_pool:
            eng = cls._engine_pool.pop()
            with stopit.ThreadingTimeout(timeout) as to_ctx_mgr:
                assert to_ctx_mgr.state == to_ctx_mgr.EXECUTING
                try:
                    eng.quit()
                except Exception:
                    # already closed or not working
                    pass
            if not to_ctx_mgr:
                warn('engine {} not responding for {} seconds'.format(str(eng), timeout))

    def __enter__(self):
        self.engine = self.pop()
        return self.engine

    def __exit__(self, exc_type, exc_value, traceback):
        self.push(self.engine)
        self.engine = None
        return exc_type is None


class MatlabMethod(Actor):
    def __init__(self, method_name, inports=(), outports='result'):
        self.method_name = method_name

        super(MatlabMethod, self).__init__(name=self.method_name)

        if isinstance(inports, six.string_types):
            inports = (inports,)
        for p in inports:
            self.inports.append(p)

        if isinstance(outports, six.string_types):
            outports = (outports,)
        for p in outports:
            self.outports.append(p)

        self._engine = None

    def run(self):
        args = (port.pop() for port in self.inports)
        func_res = self.__call__(*args)

        if len(self.outports) == 1:
            func_res = (func_res,)

        res = {}
        for name, value in zip(self.outports.keys(), func_res):
            res[name] = value
        return res

    def __call__(self, *args, **kwargs):
        with EngineManager() as engine:
            mfunc = getattr(engine, self.method_name)
            res = mfunc(*args, **kwargs)
        return res



from __future__ import absolute_import, division, print_function, unicode_literals
from collections import OrderedDict as _OrderedDict
import warnings
import six
import warnings as _warnings
import os as _os
import sys
from tempfile import mkdtemp

try:
    if six.PY3:
        from threading import get_ident as _get_ident
    else:
        from thread import get_ident as _get_ident
except ImportError:
    from dummy_threading import get_ident as _get_ident
# import a pickling module
import pickle
my_pickle = pickle
# try dill
try:
    import dill
    _IS_DILL = True
    my_pickle = dill
except ImportError:
    _IS_DILL = False
# cludpickle is default
try:
    import cloudpickle
    _IS_CLOUDPICKLE = True
    my_pickle = cloudpickle
except ImportError:
    _IS_CLOUDPICKLE = False


class ListDict(_OrderedDict):
    """Ordered dict with insert methods

    From https://gist.github.com/jaredks/6276032
    """

    def __init__(self, *args, **kwds):
        try:
            self.__insertions_running
        except AttributeError:
            self.__insertions_running = {}
        super(ListDict, self).__init__(*args, **kwds)

    def __setitem__(self, key, value):
        if _get_ident() in self.__insertions_running:
            self.__insertions_running[_get_ident()] = key, value
        else:
            super(ListDict, self).__setitem__(key, value)

    def __insertion(self, link_prev, key_value):
        self.__insertions_running[_get_ident()] = 1
        self.__setitem__(*key_value)
        key, value = self.__insertions_running.pop(_get_ident())
        if link_prev[2] != key:
            if key in self:
                del self[key]
            link_next = link_prev[1]
            self._OrderedDict__map[key] = link_prev[1] = link_next[0] = [
                link_prev, link_next, key
            ]
        dict.__setitem__(self, key, value)

    def insert_after(self, existing_key, key_value):
        self.__insertion(self._OrderedDict__map[existing_key], key_value)

    def insert_before(self, existing_key, key_value):
        self.__insertion(self._OrderedDict__map[existing_key][0], key_value)


def deprecated(func):
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emmitted
    when the function is used."""

    def new_func(*args, **kwargs):
        warnings.simplefilter('always', DeprecationWarning)  # turn off filter
        warnings.warn("Call to deprecated function {}.".format(func.__name__),
                      category=DeprecationWarning,
                      stacklevel=2)
        warnings.simplefilter('default', DeprecationWarning)  # reset filter
        return func(*args, **kwargs)

    new_func.__name__ = func.__name__
    new_func.__doc__ = func.__doc__
    new_func.__dict__.update(func.__dict__)
    return new_func


class ConstructorWrapper(object):
    """This can be used as deferred construction call."""

    def __init__(self, klass, *args, **kwargs):
        self.klass = klass
        self.args = args
        self.kwargs = kwargs

    def __call__(self):
        return self.klass(*self.args, **self.kwargs)


class TemporaryDirectory(object):
    """Create and return a temporary directory.  This has the same
    behavior as mkdtemp but can be used as a context manager.  For
    example:

        with TemporaryDirectory() as tmpdir:
            ...

    Upon exiting the context, the directory and everything contained
    in it are removed.
    """

    def __init__(self, suffix="", prefix="tmp", dir=None):
        self._closed = False
        self.name = None  # Handle mkdtemp raising an exception
        self.name = mkdtemp(suffix, prefix, dir)

    def __repr__(self):
        return "<{} {!r}>".format(self.__class__.__name__, self.name)

    def __enter__(self):
        return self.name

    def cleanup(self, _warn=False):
        if self.name and not self._closed:
            try:
                self._rmtree(self.name)
            except (TypeError, AttributeError) as ex:
                # Issue #10188: Emit a warning on stderr
                # if the directory could not be cleaned
                # up due to missing globals
                if "None" not in str(ex):
                    raise
                print("ERROR: {!r} while cleaning up {!r}".format(ex, self,),
                      file=sys.stderr)
                return
            self._closed = True
            if _warn:
                self._warn("Implicitly cleaning up {!r}".format(self))

    def __exit__(self, exc, value, tb):
        self.cleanup()

    def __del__(self):
        # Issue a ResourceWarning if implicit cleanup needed
        self.cleanup(_warn=True)

    # XXX (ncoghlan): The following code attempts to make
    # this class tolerant of the module nulling out process
    # that happens during CPython interpreter shutdown
    # Alas, it doesn't actually manage it. See issue #10188
    _listdir = staticmethod(_os.listdir)
    _path_join = staticmethod(_os.path.join)
    _isdir = staticmethod(_os.path.isdir)
    _islink = staticmethod(_os.path.islink)
    _remove = staticmethod(_os.remove)
    _rmdir = staticmethod(_os.rmdir)
    _warn = _warnings.warn

    def _rmtree(self, path):
        # Essentially a stripped down version of shutil.rmtree.  We can't
        # use globals because they may be None'ed out at shutdown.
        for name in self._listdir(path):
            fullname = self._path_join(path, name)
            try:
                isdir = self._isdir(fullname) and not self._islink(fullname)
            except OSError:
                isdir = False
            if isdir:
                self._rmtree(fullname)
            else:
                try:
                    self._remove(fullname)
                except OSError:
                    pass
        try:
            self._rmdir(path)
        except OSError:
            pass


def enum(*sequential, **named):
    """Handy way to fake an enumerated type in Python
    http://stackoverflow.com/questions/36932/how-can-i-represent-an-enum-in-python
    """
    enums = dict(zip(sequential, range(len(sequential))), **named)
    return type(str('Enum'), (), enums)


MPI_TAGS = enum('READY', 'DONE', 'EXIT', 'START')


def dumps(obj):
    return my_pickle.dumps(obj)


def loads(obj):
    return my_pickle.loads(obj)


def dump(obj, file):
    if isinstance(file, six.string_types):
        file = open(file, 'wb')
    return my_pickle.dump(obj, file)


def load(file):
    if isinstance(file, six.string_types):
        file = open(file, 'rb')
    return my_pickle.load(file)


def abstractmethod(method):
    def default_abstract_method(*args, **kwargs):
        raise NotImplementedError('call to abstract method ' + repr(method))

    default_abstract_method.__name__ = method.__name__

    return default_abstract_method



#!/usr/bin/env python
from mpi4py import MPI
import numpy
import sys
import os
import cloudpickle
from time import sleep


def spam(x):
    return 4.0 / (1.0 + x**2)


spam_p = cloudpickle.dumps(spam)

comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()

print('master rank {}, size {}, pid {}'.format(rank, size, os.getpid()))

# comm = MPI.COMM_SELF.Spawn(sys.executable,
#                            args=['mpi_worker.py'],
#                            maxprocs=5)

print('root {}'.format(MPI.ROOT))
sleep(2)
print('bcast 1')
comm.bcast(spam_p, root=MPI.ROOT)

N = numpy.array(100, 'i')
comm.Bcast([N, MPI.INT], root=MPI.ROOT)
PI = numpy.array(0.0, 'd')
comm.Reduce(None, [PI, MPI.DOUBLE],
            op=MPI.SUM, root=MPI.ROOT)
print(PI)

comm.Disconnect()



from __future__ import absolute_import, division, print_function

import itertools

from ..components import Actor
from ..schedulers import _ActorRunner, ThreadedScheduler
from future.builtins import super


class Concat(Actor):

    _system_actor = True

    def __init__(self, n_ports, name='concat'):
        super().__init__(name=name)
        for i in range(n_ports):
            self.inports.append('in_{}'.format(i))
        self.outports.append('out')
        self._last_connected = -1

    def get_run_args(self):
        return (), {}

    def run(self, *args, **kwargs):
        res = tuple((port.pop() for port in self.inports))
        return {'out': res}

    def connect_input(self, port, i=None):
        """
        Connects i-th input. If i is None, the last value + 1 is used.

        Args:
            port (OutPort): output port to connect
            i (int): input port index
        """

        if i is None:
            i = self._last_connected + 1
        self._last_connected = i
        self.inports['in_{}'.format(i)].connect(port)


class MultiConcat(Actor):

    _system_actor = True

    def __init__(self, name='multiconcat'):
        super().__init__(name=name)
        self._port_names = []
        self._last_connected = -1

    def get_run_args(self):
        return (), {}

    def run(self, *args, **kwargs):
        res = {}
        for inport, outport in zip(self.inports, itertools.cycle(self.outports)):
            res.setdefault(outport.name, [])
            res[outport.name].append(inport.pop())
        return res

    def add_and_connect(self, actor):
        if self._last_connected > -1:
            assert self._port_names == [port.name for port in actor.outports]
        self._last_connected += 1
        i = self._last_connected
        for port in actor.outports:
            self.inports.append('{}_{}'.format(port.name, i))
            self.inports.at(-1).connect(port)
            if i == 0:
                self.outports.append(port.name)
                self._port_names.append(port.name)


class Map(Actor):
    """
    Maps a given actor on input token elements.

    The mapped actor is assumed to have a single input and a single output port.

    Args:
        actor (Actor): Actor class to be mapped to inputs
        args (list): positional arguments for actor.__init__
        kwargs (dict): keyword arguments for actor.__init__
        scheduler (Scheduler): scheduler to use, default is not to change the scheduler
        name (string): actor name

    Ports:
        in (iterable): contains items to be passed to the mapped actor
        out (tuple): items after applying the map actor
    """

    _system_actor = True

    def __init__(self, actor_class, args=(), kwargs={}, scheduler=None, name='map'):
        super().__init__(name=name)
        self.actor_class = actor_class
        self.actor_args = args
        self.actor_kwargs = kwargs
        self.map_scheduler = scheduler
        # get port names from an actor instance
        actor = self.actor_class(*self.actor_args, **self.actor_kwargs)
        for pname in actor.inports.keys():
            self.inports.append(pname)
        for pname in actor.outports.keys():
            self.outports.append(pname)

    def get_run_args(self):
        return (), {}

    def run(self, *args, **kwargs):
        # run is not a classfunction for Map
        # bacause it needs to change the workflow
        if self.map_scheduler is None:
            # self.schduler is set by the calling scheduler
            map_scheduler = self.scheduler.copy()
        elif isinstance(self.map_scheduler, (_ActorRunner, ThreadedScheduler)):
            # we need a copy of an existing scheduler
            map_scheduler = self.map_scheduler.copy()
        else:
            # in this case, we assume self.map_scheduler is a class
            map_scheduler = self.map_scheduler()
        # destinations = [port for port in self.outports['out'].connections]
        # disconnect the output port
        # for port in destinations:
        #     self.outports['out'].disconnect(port)
        # create the map actors, connect and put inputs
        map_actors = []
        concat_actor = MultiConcat()
        # items will iterate over all ports inputs, i.e. will contain n-th actor input
        # TODO ensure equal input lengths
        for items in zip(*(port.pop() for port in self.inports)):
            # get actor instance
            actor = self.actor_class(*self.actor_args, **self.actor_kwargs)
            map_actors.append(actor)
            concat_actor.add_and_connect(actor)
            # put the input data item
            for port, value in zip(actor.inports, items):
                # value is a single input item of the port
                map_scheduler.put_value(port, value)
        # run the mapping sub-workflows
        # map_workflow = concat_actor.get_workflow()
        # result = map_scheduler.run_workflow(map_workflow)
        map_scheduler.execute()

        result = {port.name: port.pop() for port in concat_actor.outports}
        return result


class PassWID(Actor):
    """
    Pass input argument with hostname and process ID attached
    """
    def __init__(self, name='passwid'):
        super().__init__(name=name)
        self.inports.append('inp')
        self.outports.append('out')

    def get_run_args(self):
        return (self.inports['inp'].pop(), ), {}

    @staticmethod
    def run(*args, **kwargs):
        from os import getpid
        from socket import gethostname
        return {'out':
                    {'inp': args[0],
                     'host': gethostname(),
                     'pid': getpid()}}



#!/usr/bin/env python
from setuptools import setup, find_packages
from _wowp_setup import get_version

setup(
    name='wowp',
    version=get_version(),
    packages=find_packages(),
    scripts=['scripts/wowp', 'scripts/wowp_on_slurm'],
    license='MIT',
    description='Data-flow-actors-based workflow framework',
    long_description=open('README.txt').read(),
    author='Jakub Urban, Jan Pipek',
    author_email='coobas@gmail.com',
    url='http://pythonic.eu/wowp/',
    install_requires=[
        'decorator',
        'future',
        'networkx',
        'nose',
        'six',
        'click',
    ],
    extras_require={'parallel': ['ipyparallel', 'mpi4py']},
    entry_points={
        # 'console_scripts' : [
        #    'wowp = wowp:function_that_does_it_all'
        # ]
    },
    classifiers=[
        "Programming Language :: Python :: 2.7", "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5", "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License", "Operating System :: OS Independent",
        "Intended Audience :: Developers", "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research", "Topic :: Scientific/Engineering",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
    test_suite='nose.collector',
    tests_require='nose>=1.0')



from . import Actor
from ..schedulers import LinearizedScheduler


class GeneratorActor(Actor):
    class PseudoDict(object):
        def __init__(self, iterator):
            self.iterator = iterator

        def items(self):
            return self.iterator

    def run(self):
        return GeneratorActor.PseudoDict(self.iterate())

    def iterate(self):
        raise NotImplementedError(
            "It is necessary to implement iterate method that yields pairs key, value")


class LineReader(GeneratorActor):
    """Sequentially put all lines in a file."""

    def __init__(self, name="line_reader", inport_name="path", outport_name="line"):
        Actor.__init__(self, name=name)
        self.inports.append(inport_name)
        self.outports.append(outport_name)
        self.inport_name = inport_name
        self.outport_name = outport_name

    def iterate(self):
        path = self.inports[self.inport_name].pop()
        with open(path, "rt") as f:
            for line in f:
                yield self.outport_name, line.strip()


class IteratorActor(GeneratorActor):
    def __init__(self, name="iterator", inport_name="collection", outport_name="item"):
        Actor.__init__(self, name=name)
        self.inports.append(inport_name)
        self.outports.append(outport_name)
        self.inport_name = inport_name
        self.outport_name = outport_name

    def iterate(self):
        collection = self.inports[self.inport_name].pop()
        for item in collection:
            yield self.outport_name, item


class Splitter(Actor):
    def __init__(self, name="splitter", inport_name="in", multiplicity=2):
        import itertools
        Actor.__init__(self, name=name)
        self.inport_name = inport_name
        self.multiplicity = multiplicity

        self.inports.append(inport_name)
        for i in range(1, multiplicity + 1):
            self.outports.append("%s_%d" % (inport_name, i))

        self._outports_cycle = itertools.cycle(range(1, multiplicity + 1))

    def run(self):
        value = self.inports[self.inport_name].pop()
        i = next(self._outports_cycle)
        outport = "%s_%d" % (self.inport_name, i)
        return {outport: value}

        # TODO: Add SequentialMerger
        # TODO: Add RandomMerger


class Chain(Actor):
    """Chain of actors.

    Each of the chain link has to have just one input & output port (to be changed)
    """

    def __init__(self, name, actor_generators, **kwargs):
        """

        :param name:
        :param actor_generators: iterable of actor classes or generators
        :param kwargs:
        :return:
        """
        super(Chain, self).__init__(name=name)
        self.actors = []
        if len(actor_generators) < 1:
            raise RuntimeError("Chain needs at least one item.")
        self.inports.append("inp")
        self.outports.append("out")
        self.actor_generators = actor_generators

    def get_run_args(self):
        return (self.inports["inp"].pop(), ), {"generators": self.actor_generators}

    @staticmethod
    def run(*args, **kwargs):
        input = args[0]
        actor_generators = kwargs.pop("generators")
        actors = []
        for generator in actor_generators:
            actor = generator()
            assert len(actor.inports) == 1
            assert len(actor.outports) == 1
            if actors:
                inport = actor.inports.at(0)
                inport += actors[-1].outports.at(0)
            actors.append(actor)

        scheduler = LinearizedScheduler()
        scheduler.put_value(actors[0].inports.at(0), input)
        scheduler.execute()
        return {"out": actors[-1].outports.at(0).pop()}
